// Package httpcli is http request client, which only supports return json format.
package httpcli

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"gitlab.wanzhuangkj.com/tush/xpkg/httpcli/entity"
	"io"
	"net/http"
	"net/url"
	"strings"
	"time"

	"github.com/duke-git/lancet/v2/retry"
	"gitlab.wanzhuangkj.com/tush/xpkg/gin/ctxUtils"
	"gitlab.wanzhuangkj.com/tush/xpkg/logger"
)

const defaultTimeout = 30 * time.Second

type Request struct {
	url         string
	method      string
	headers     map[string]string
	params      map[string]interface{} // parameters after URL
	reqBody     []byte                 // Body data
	reqBodyJSON interface{}            // 传入对象，会自动json marshal
	timeout     time.Duration          // Client timeout
	retryCount  uint

	request  *http.Request
	response *Response
	err      error
}

// Response HTTP response
type Response struct {
	body       []byte
	StatusCode int
	Status     string
	Header     http.Header
}

// -----------------------------------  Request way 1 -----------------------------------

// New create a new Request
func New() *Request {
	return &Request{
		response: &Response{},
	}
}

// Reset set all fields to default value, use at pool
func (req *Request) Reset() {
	req.params = nil
	req.reqBody = nil
	req.reqBodyJSON = nil
	req.timeout = 0
	req.headers = nil

	req.response = nil
	req.method = ""
	req.err = nil
}

// SetURL set URL
func (req *Request) SetURL(path string) *Request {
	req.url = path
	return req
}

// SetParams parameters after setting the URL
func (req *Request) SetParams(params map[string]interface{}) *Request {
	if req.params == nil {
		req.params = params
	} else {
		for k, v := range params {
			req.params[k] = v
		}
	}
	return req
}

// SetParam parameters after setting the URL
func (req *Request) SetParam(k string, v interface{}) *Request {
	if req.params == nil {
		req.params = make(map[string]interface{})
	}
	req.params[k] = v
	return req
}

// SetBody set body data, support string and []byte, if it is not string, it will be json marshal.
func (req *Request) SetBody(body interface{}) *Request {
	switch v := body.(type) {
	case string:
		req.reqBody = []byte(v)
	case []byte:
		req.reqBody = v
	default:
		req.reqBodyJSON = body
	}
	return req
}

// SetTimeout set timeout
func (req *Request) SetTimeout(t time.Duration) *Request {
	req.timeout = t
	return req
}
func (req *Request) SetRetry(count uint) *Request {
	req.retryCount = count
	return req
}

// SetContentType set ContentType
func (req *Request) SetContentType(a string) *Request {
	req.SetHeader("Content-Type", a)
	return req
}

// SetHeader set the value of the request header
func (req *Request) SetHeader(k, v string) *Request {
	if req.headers == nil {
		req.headers = make(map[string]string)
	}
	req.headers[k] = v
	return req
}

// SetHeaders set the value of Request Headers
func (req *Request) SetHeaders(headers map[string]string) *Request {
	if req.headers == nil {
		req.headers = make(map[string]string)
	}
	for k, v := range headers {
		req.headers[k] = v
	}
	return req
}

// GET send a GET request
func (req *Request) GET(ctx context.Context) *Request {
	req.method = http.MethodGet
	req.pull(ctx)
	return req
}

// DELETE send a DELETE request
func (req *Request) DELETE(ctx context.Context) *Request {
	req.method = http.MethodDelete
	req.pull(ctx)
	return req
}

// POST send a POST request
func (req *Request) POST(ctx context.Context) *Request {
	req.method = http.MethodPost
	req.push(ctx)
	return req
}

// PUT send a PUT request
func (req *Request) PUT(ctx context.Context) *Request {
	req.method = http.MethodPut
	req.push(ctx)
	return req
}

// PATCH send PATCH requests
func (req *Request) PATCH(ctx context.Context) *Request {
	req.method = http.MethodPatch
	req.push(ctx)
	return req
}

// Do a request
func (req *Request) Do(ctx context.Context, method string, data interface{}) *Request {
	req.method = method
	switch method {
	case http.MethodGet, http.MethodDelete:
		if data != nil {
			if params, ok := data.(map[string]interface{}); ok { //nolint
				req.SetParams(params)
			} else {
				req.err = errors.New("params is not a map[string]interface{}")
				return req
			}
		}
		req.pull(ctx)
	case http.MethodPost, http.MethodPut, http.MethodPatch:
		if data != nil {
			req.SetBody(data)
		}
		req.push(ctx)
	default:
		req.err = errors.New("unknow method " + method)
	}
	return req
}

func (req *Request) pull(ctx context.Context) {
	val := ""
	if len(req.params) > 0 {
		values := url.Values{}
		for k, v := range req.params {
			values.Add(k, fmt.Sprintf("%v", v))
		}
		val += values.Encode()
	}

	if val != "" {
		if strings.Contains(req.url, "?") {
			req.url += "&" + val
		} else {
			req.url += "?" + val
		}
	}
	req.send(ctx)
}

func (req *Request) push(ctx context.Context) {
	if req.reqBodyJSON != nil {
		body, err := json.Marshal(req.reqBodyJSON)
		if err != nil {
			req.err = err
			return
		}
		req.reqBody = body
	}
	req.send(ctx)
}

func (req *Request) send(ctx context.Context) {
	bodyBuf := bytes.NewBuffer(req.reqBody)
	var request *http.Request
	request, req.err = http.NewRequest(req.method, req.url, bodyBuf)
	if req.err != nil {
		return
	}
	if req.reqBody != nil {
		if c, _ := ctxUtils.GetGinCtx(ctx); c != nil {
			entity.SetCopyApiReq(c, bodyBuf)
		}
	}
	if req.headers != nil {
		for k, v := range req.headers {
			request.Header.Add(k, v)
		}
	}
	request.Header.Add(ctxUtils.HeaderXRequestIDKey, ctxUtils.CtxRequestID(ctx))
	if req.timeout < 1 {
		req.timeout = defaultTimeout
	}
	ctx, cancel := context.WithTimeout(ctx, req.timeout)
	defer cancel()
	req.request.WithContext(ctx)

	logger.Info("[httpCli] req",
		logger.Any("method", req.method),
		logger.Any("url", req.url),
		logger.Any("header", request.Header),
		logger.Any("body", bodyBuf.String()),
		ctxUtils.CtxTraceIDField(ctx))

	req.request = request

	if req.retryCount > 0 {
		_ = retry.Retry(func() error {
			req.pushDo(ctx)
			return req.err
		},
			retry.RetryTimes(req.retryCount),
			retry.RetryWithLinearBackoff(8*time.Second),
			retry.Context(ctx),
		)
	} else {
		req.pushDo(ctx)
	}

	var rspHeader map[string][]string
	var rspBodyStr string
	if req.response != nil {
		rspHeader = req.response.Header
		if len(req.response.body) > 0 {
			rspBodyStr = bytes.NewBuffer(req.response.body).String()
		}
	}
	logger.Info("[httpCli] rsp",
		logger.Any("method", req.method),
		logger.Any("status", req.response.Status),
		logger.Any("url", req.url),
		logger.Any("header", rspHeader),
		logger.Any("body", rspBodyStr),
		ctxUtils.CtxTraceIDField(ctx))
}

func (req *Request) pushDo(ctx context.Context) {
	client := http.Client{Timeout: req.timeout}
	response, e := client.Do(req.request)
	if e != nil {
		req.request = nil
		req.err = e
		return
	}
	if response != nil {
		req.response.Status = response.Status
		req.response.Header = response.Header
		ctxUtils.SetApiCost(ctx, req.response.Header)
		req.response.StatusCode = response.StatusCode
		if response.Body != nil {
			defer response.Body.Close()
			req.response.body, req.err = io.ReadAll(response.Body)
			if req.err != nil {
				req.request = nil
				return
			}
			if len(req.response.body) > 0 {
				if c, _ := ctxUtils.GetGinCtx(ctx); c != nil {
					entity.SetCopyApiRsp(c, bytes.NewBuffer(req.response.body))
				}
				//req.response.Body = io.NopCloser(bytes.NewBuffer(req.rspBody))
			}
		}
	}
	req.request = nil
}

// -----------------------------------  Response -----------------------------------

func (req *Request) Err() error {
	return req.err
}

// BodyString returns the body data of the HttpResponse
func (req *Request) BodyString() string {
	if len(req.response.body) == 0 {
		return ""
	}
	return string(req.response.body)
}
func (req *Request) Response() *Response {
	return req.response
}

func (req *Request) ReadRspBody() ([]byte, error) {
	if req.err != nil {
		return []byte{}, req.err
	}

	if req.response.body == nil {
		return []byte{}, errors.New("nil")
	}

	return req.response.body, nil
}

// BindJSON parses the response's body as JSON
func (req *Request) BindJSON(v interface{}) *Request {
	if req.err != nil {
		return req
	}
	if err := json.Unmarshal(req.response.body, v); err != nil {
		req.err = err
	}
	return req
}

// -----------------------------------  Request way 2 -----------------------------------

// Option set options.
type Option func(*options)

type options struct {
	params  map[string]interface{}
	headers map[string]string
	timeout time.Duration
}

func (o *options) apply(opts ...Option) {
	for _, opt := range opts {
		opt(o)
	}
}

func defaultOptions() *options {
	return &options{}
}

// WithParams set params
func WithParams(params map[string]interface{}) Option {
	return func(o *options) {
		o.params = params
	}
}

// WithHeaders set headers
func WithHeaders(headers map[string]string) Option {
	return func(o *options) {
		o.headers = headers
	}
}

// WithTimeout set timeout
func WithTimeout(t time.Duration) Option {
	return func(o *options) {
		o.timeout = t
	}
}

//
//// Get request, return custom json format
//func Get(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return gDo(ctx, "GET", result, urlStr, o.params, o.headers, o.timeout)
//}
//
//// Delete request, return custom json format
//func Delete(ctx context.Context, result interface{}, urlStr string, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return gDo(ctx, "DELETE", result, urlStr, o.params, o.headers, o.timeout)
//}
//
//// Post request, return custom json format
//func Post(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return do(ctx, "POST", result, urlStr, body, o.params, o.headers, o.timeout)
//}
//
//// Put request, return custom json format
//func Put(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return do(ctx, "PUT", result, urlStr, body, o.params, o.headers, o.timeout)
//}
//
//// Patch request, return custom json format
//func Patch(ctx context.Context, result interface{}, urlStr string, body interface{}, opts ...Option) error {
//	o := defaultOptions()
//	o.apply(opts...)
//	return do(ctx, "PATCH", result, urlStr, body, o.params, o.headers, o.timeout)
//}
//
//var requestErr = func(err error) error { return fmt.Errorf("request error, err=%v", err) }
//var jsonParseErr = func(err error) error { return fmt.Errorf("json parsing error, err=%v", err) }
//var notOKErr = func(resp *Response) error {
//	body, err := resp.ReadBody()
//	if err != nil {
//		return err
//	}
//	if len(body) > 500 {
//		body = append(body[:500], []byte(" ......")...)
//	}
//	return fmt.Errorf("statusCode=%d, body=%s", resp.StatusCode, body)
//}
//
//func do(ctx context.Context, method string, result interface{}, urlStr string, body interface{}, params KV, headers map[string]string, timeout time.Duration) error {
//	if result == nil {
//		return fmt.Errorf("'result' can not be nil")
//	}
//
//	req := &Request{}
//	req.SetURL(urlStr)
//	req.SetContentType("application/json")
//	req.SetParams(params)
//	req.SetHeaders(headers)
//	req.SetBody(body)
//	req.SetTimeout(timeout)
//
//	var resp *Response
//	var err error
//	switch method {
//	case "POST":
//		resp, err = req.POST(ctx)
//	case "PUT":
//		resp, err = req.PUT(ctx)
//	case "PATCH":
//		resp, err = req.PATCH(ctx)
//	}
//	if err != nil {
//		return requestErr(err)
//	}
//	defer resp.Body.Close() //nolint
//
//	if resp.StatusCode != 200 {
//		return notOKErr(resp)
//	}
//
//	err = resp.BindJSON(result)
//	if err != nil {
//		return jsonParseErr(err)
//	}
//
//	return nil
//}

//
//func gDo(ctx context.Context, method string, result interface{}, urlStr string, params KV, headers map[string]string, timeout time.Duration) error {
//	req := &Request{}
//	req.SetURL(urlStr)
//	req.SetParams(params)
//	req.SetHeaders(headers)
//	req.SetTimeout(timeout)
//
//	var resp *Response
//	var err error
//	switch method {
//	case "GET":
//		resp, err = req.GET(ctx)
//	case "DELETE":
//		resp, err = req.DELETE(ctx)
//	}
//	if err != nil {
//		return requestErr(err)
//	}
//	defer resp.Body.Close() //nolint
//
//	if resp.StatusCode != 200 {
//		return notOKErr(resp)
//	}
//
//	err = resp.BindJSON(result)
//	if err != nil {
//		return jsonParseErr(err)
//	}
//
//	return nil
//}

// StdResult standard return data
type StdResult struct {
	Code int         `json:"code"`
	Msg  string      `json:"msg"`
	Data interface{} `json:"data,omitempty"`
}

// StdResult standard return data
type StdResult2 struct {
	Code int         `json:"code"`
	Msg  string      `json:"message"`
	Data interface{} `json:"data,omitempty"`
}

// KV string:interface{}
type KV = map[string]interface{}
