// Package query is a library of custom condition queries, support for complex conditional paging queries.
package query

import (
	"fmt"
	"strings"

	"gitlab.wanzhuangkj.com/tush/xpkg/utils/xtime"
	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"
)

const (
	// Eq equal
	Eq = "eq"
	// Neq not equal
	Neq = "neq"
	// Gt greater than
	Gt = "gt"
	// Gte greater than or equal
	Gte = "gte"
	// Lt less than
	Lt = "lt"
	// Lte less than or equal
	Lte = "lte"
	// Like fuzzy lookup
	Like = "like"
	// In include
	In = "in"

	// AND logic and
	AND string = "and"
	// OR logic or
	OR string = "or"
)

var expMap = map[string]string{
	Eq:   " = ",
	Neq:  " <> ",
	Gt:   " > ",
	Gte:  " >= ",
	Lt:   " < ",
	Lte:  " <= ",
	Like: " LIKE ",
	In:   " IN ",

	"=":  " = ",
	"!=": " <> ",
	">":  " > ",
	">=": " >= ",
	"<":  " < ",
	"<=": " <= ",
}

var logicMap = map[string]string{
	AND: " AND ",
	OR:  " OR ",

	"&":   " AND ",
	"&&":  " AND ",
	"|":   " OR ",
	"||":  " OR ",
	"AND": " AND ",
	"OR":  " OR ",
}

type IPagination interface {
	GetOffset() int
	GetLimit() int
}

type Pagination struct {
	PageIndex int    `json:"pageIndex" form:"pageIndex" binding:"" example:"1"`                      // 1-based
	PageSize  int    `json:"pageSize" form:"pageSize" binding:""  example:"10"`                      // 1-based
	Sort      string `json:"sort,omitempty" form:"sort" binding:"" example:"-id" query:"type:order"` //排序字段，负号表示降序，多个字段排序中间用,(英文逗号)分隔
	Unscope   bool   `json:"-" form:"-" query:"-"`
}

func (x Pagination) Unscoped() bool {
	return x.Unscope
}

type CreatedAtPeriod struct {
	CreatedAtBegin xtime.DateTime `json:"createdAtBegin" form:"createdAtBegin" gorm:"column:created_at;type:datetime;comment:创建时间" query:"type:gte;column:created_at" example:"2025-01-01 12:20:26"` //开始日期起始 (>=)
	CreatedAtEnd   xtime.DateTime `json:"createdAtEnd" form:"createdAtEnd" gorm:"column:created_at;type:datetime;comment:创建时间"  query:"type:lt;column:created_at" example:"2025-02-01 12:20:26"`     //结束日期截止 (<)
}

func (x Pagination) GetOffset() int {
	if x.PageIndex <= 0 {
		x.PageIndex = 1
	}
	return (x.PageIndex - 1) * x.GetPageSize()
}

func (x Pagination) GetPageSize() int {
	if x.PageSize <= 0 {
		x.PageSize = 10
	}
	return x.PageSize
}

func (x Pagination) GetPageIndex() int {
	if x.PageIndex <= 0 {
		x.PageIndex = 1
	}
	return x.PageIndex
}

func (x Pagination) GetLimit() int {
	return x.GetPageSize()
}

func (x Pagination) GetSort() string {
	if x.Sort == "" {
		x.Sort = "id desc"
	}
	return x.Sort
}

// Params query parameters
type Params struct {
	PageIndex int      `json:"pageIndex" form:"pageIndex" binding:"gte=1" validate:"required" example:"1"` // 1-based
	PageSize  int      `json:"pageSize" form:"pageSize" binding:"gte=1" validate:"required" example:"10"`  // 1-based
	Sort      string   `json:"sort,omitempty" form:"sort" binding:"" validate:"required" example:"id"`
	Columns   []Column `json:"columns,omitempty" form:"columns"` // not required
}

// Column query info
type Column struct {
	Name  string      `json:"name" form:"name" validate:"required" example:"id"` // column name eg id
	Exp   string      `json:"exp" form:"exp" validate:"required" example:"="`    // expressions, default value is "=", support =, !=, >, >=, <, <=, like, in
	Value interface{} `json:"value" form:"value" validate:"required"`            // column value eg 102
	Logic string      `json:"logic" form:"logic" example:"and"`                  // logical type, defaults to and when the value is null, with &(and), ||(or)
}

func (c *Column) checkValid() error {
	if c.Name == "" {
		return xerror.New("field 'name' cannot be empty")
	}
	if c.Value == nil {
		return xerror.New("field 'value' cannot be nil")
	}
	return nil
}

// converting ExpType to sql expressions and LogicType to sql using characters
func (c *Column) convert() error {
	if c.Exp == "" {
		c.Exp = Eq
	}
	if v, ok := expMap[strings.ToLower(c.Exp)]; ok { //nolint
		c.Exp = v
		if c.Exp == " LIKE " {
			c.Value = fmt.Sprintf("%%%v%%", c.Value)
		}
		if c.Exp == " IN " {
			val, ok := c.Value.(string)
			if !ok {
				return xerror.Newf("invalid value type '%s'", c.Value)
			}
			iVal := []interface{}{}
			ss := strings.Split(val, ",")
			for _, s := range ss {
				iVal = append(iVal, s)
			}
			c.Value = iVal
		}
	} else {
		return xerror.Newf("unknown exp type '%s'", c.Exp)
	}

	if c.Logic == "" {
		c.Logic = AND
	}
	if v, ok := logicMap[strings.ToLower(c.Logic)]; ok { //nolint
		c.Logic = v
	} else {
		return xerror.Newf("unknown logic type '%s'", c.Logic)
	}

	return nil
}

// ConvertToPage converted to page
func (p *Params) ConvertToPage() (order string, limit int, offset int) { //nolint
	page := NewPage(p.PageIndex, p.PageSize, p.Sort)
	order = page.sort
	limit = page.pageSize
	offset = page.pageIndex * page.pageSize
	return //nolint
}

// ConvertToGormConditions conversion to gorm-compliant parameters based on the Columns parameter
// ignore the logical type of the last column, whether it is a one-column or multi-column query
func (p *Params) ConvertToGormConditions() (string, []interface{}, error) {
	str := ""
	args := []interface{}{}
	l := len(p.Columns)
	if l == 0 {
		return "", nil, nil
	}

	isUseIN := true
	if l == 1 {
		isUseIN = false
	}
	field := p.Columns[0].Name

	for i, column := range p.Columns {
		if err := column.checkValid(); err != nil {
			return "", nil, err
		}

		err := column.convert()
		if err != nil {
			return "", nil, err
		}

		symbol := "?"
		if column.Exp == " IN " {
			symbol = "(?)"
		}
		if i == l-1 { // ignore the logical type of the last column
			str += column.Name + column.Exp + symbol
		} else {
			str += column.Name + column.Exp + symbol + column.Logic
		}
		args = append(args, column.Value)

		// when multiple columns are the same, determine whether the use of IN
		if isUseIN {
			if field != column.Name {
				isUseIN = false
				continue
			}
			if column.Exp != expMap[Eq] {
				isUseIN = false
			}
		}
	}

	if isUseIN {
		str = field + " IN (?)"
		args = []interface{}{args}
	}

	return str, args, nil
}

// Conditions query conditions
type Conditions struct {
	Columns []Column `json:"columns" form:"columns" binding:"min=1"` // columns info
	Sort    string   `json:"sort,omitempty" form:"sort" binding:"" validate:"required" example:"id"`
}

// CheckValid check valid
func (c *Conditions) CheckValid() error {
	if len(c.Columns) == 0 {
		return xerror.New("field 'columns' cannot be empty")
	}
	for _, column := range c.Columns {
		err := column.checkValid()
		if err != nil {
			return err
		}
		if column.Exp != "" {
			if _, ok := expMap[column.Exp]; !ok {
				return xerror.Newf("unknown exp type '%s'", column.Exp)
			}
		}
		if column.Logic != "" {
			if _, ok := logicMap[column.Logic]; !ok {
				return xerror.Newf("unknown logic type '%s'", column.Logic)
			}
		}
	}

	return nil
}

// ConvertToGorm conversion to gorm-compliant parameters based on the Columns parameter
// ignore the logical type of the last column, whether it is a one-column or multi-column query
func (c *Conditions) ConvertToGorm() (string, []interface{}, error) {
	p := &Params{Columns: c.Columns}
	return p.ConvertToGormConditions()
}
